@@ -1,3 +1,5 @@
+from __future__ import unicode_literals
+
 import re
 import io
 import json
@@ -7,6 +9,7 @@ import six
 import ssl
 from six.moves import queue
 
+import socket
 import websocket
 from .messages import *
 from .utils import logger
@@ -50,13 +53,21 @@ class WampInvokeWrapper(threading.Thread):
             ))
         except Exception as ex:
             error_uri = self.client.get_full_uri('error.invoke.failure')
+
+            exargs = ["Call failed: {}".format(ex)]
+            try:
+                json.dumps(ex.args) # Just testing
+                exargs += list(ex.args)
+            except TypeError as err:
+                logger.warning("Unable to serialize exception arguments: {}".format(ex))
+
             try:
                 self.client.send_message(ERROR(
                     request_code = WAMP_INVOCATION,
                     request_id = req_id,
                     details = {},
                     error = error_uri,
-                    args = [u'Call failed: {}'.format(ex)],
+                    args = exargs
                 ))
 
             # We might fail when we try to send an error message back to the
@@ -93,6 +104,9 @@ class WAMPClient(threading.Thread):
     timeout = None
     sslopt = None
     sockopt = None
+    loop_timeout = 5
+    heartbeat_timeout = 10
+    ping_interval = 3
 
     auto_reconnect = True
 
@@ -105,9 +119,13 @@ class WAMPClient(threading.Thread):
     _requests_pending = None
     _request_disconnect = None
     _request_shutdown = False
-    _loop_timeout = 0.1
     _state = STATE_DISCONNECTED
 
+    _last_ping_time = None
+    _last_pong_time = None
+    _heartbeat_thread = None
+    _stop_heartbeat = False
+
     def __init__(
                 self,
                 url='ws://localhost:8080',
@@ -117,6 +135,9 @@ class WAMPClient(threading.Thread):
                 authmethods=None,
                 authid=None,
                 timeout=10,
+                loop_timeout=5,
+                heartbeat_timeout=10,
+                ping_interval=3,
                 auto_reconnect=1,
                 sslopt=None,
                 sockopt=None,
@@ -140,6 +161,9 @@ class WAMPClient(threading.Thread):
             auto_reconnect = auto_reconnect,
             sslopt = sslopt,
             sockopt = sockopt,
+            loop_timeout = loop_timeout,
+            heartbeat_timeout = heartbeat_timeout,
+            ping_interval = ping_interval
         )
 
     def get_full_uri(self,uri):
@@ -158,8 +182,6 @@ class WAMPClient(threading.Thread):
         m = re.search('(ws+)://([\w\.]+)(:?:(\d+))?',self.url)
 
         options['subprotocols'] = ['wamp.2.json']
-        options['enable_multithread'] = True
-        options.setdefault('timeout',self._loop_timeout)
 
         # Handle the weird issue in websocket that the origin
         # port will be always http://host:port even though connection is
@@ -167,8 +189,8 @@ class WAMPClient(threading.Thread):
         # so we ensure we use http or https appropriately depending on the
         # ws or wss protocol
         if m and m.group(1).lower() == 'wss':
-            origin_port = ':'+m.group(3) if m.group(3) else ''
-            options['origin'] = u'https://{}{}'.format(m.group(2),origin_port)
+            origin_port = ':'+m.group(4) if m.group(4) else ''
+            options['origin'] = 'https://{}{}'.format(m.group(2),origin_port)
 
         # Attempt connection once unless it's autoreconnect in which
         # case we try and try again...
@@ -180,8 +202,8 @@ class WAMPClient(threading.Thread):
                 # By default if no settings are chosen we apply
                 # the looser traditional policy (makes life less
                 # secure but less excruciating on windows)
-                if options.get('sslopt') is None:
-                    options['sslopt'] = {
+                if options.get("sslopt") is None:
+                    options["sslopt"] = {
                         "cert_reqs":ssl.CERT_NONE,
                         "check_hostname": False
                     }
@@ -189,10 +211,12 @@ class WAMPClient(threading.Thread):
                 if self.sockopt:
                     options.setdefault('sockopt',self.sockopt)
 
-                self.ws = websocket.create_connection(
-                                self.url,
-                                **options
-                            )
+                self.ws = websocket.WebSocket(fire_cont_frame=options.pop("fire_cont_frame", False),
+                                    skip_utf8_validation=options.pop("skip_utf8_validation", False),
+                                    enable_multithread=True,
+                                    **options)
+                self.ws.settimeout(self.loop_timeout)
+                self.ws.connect(self.url, **options)
                 self.handle_connect()
             except Exception as ex:
                 if self.auto_reconnect:
@@ -221,6 +245,33 @@ class WAMPClient(threading.Thread):
         self._request_loop_notify_restart.notify()
         self._request_loop_notify_restart.release()
 
+    def start_heartbeat(self, event):
+        self._last_ping_time = None
+        self._last_pong_time = None
+        while not event.wait(self.ping_interval) and not self._stop_heartbeat:
+            self._last_ping_time = time.time()
+            try:
+                self.ws.ping(str(self._last_ping_time))
+            except Exception as ex:
+                raise ExWAMPConnectionError("Ping failed: %s", ex)
+        self._stop_heartbeat = False
+
+    def stop_heartbeat(self):
+        self._stop_heartbeat = True
+
+    def heartbeat(self):
+        """ starts a new thread that sends websocket ping messages to
+        the router.
+        """
+        self._stop_heartbeat = False
+        if self.ping_interval:
+            event = threading.Event()
+            thread = threading.Thread(
+                target=self.start_heartbeat, args=(event,))
+            thread.setDaemon(True)
+            thread.start()
+            self.heartbeat_thread = thread
+
     def is_disconnected(self):
         """ returns a true value if the connection is currently dead
         """
@@ -234,7 +285,8 @@ class WAMPClient(threading.Thread):
     def configure(self, **kwargs):
         for k in ('url','uri_base','realm',
                   'agent','timeout','authmethods', 'authid',
-                  'auto_reconnect', 'sslopt', 'sockopt'):
+                  'auto_reconnect', 'sslopt', 'sockopt',
+                  'loop_timeout', 'heartbeat_timeout', 'ping_interval'):
             if k in kwargs:
                 setattr(self,k,kwargs[k])
 
@@ -254,6 +306,8 @@ class WAMPClient(threading.Thread):
 
         # Then rebind all the registrations and callbacks
         # if there's a need
+        if details.details['authmethod'] != 'anonymous':
+            self.heartbeat()
         to_register = self._registered_calls
         self._registered_calls = {}
         for uri, callback in to_register.values():
@@ -280,8 +334,8 @@ class WAMPClient(threading.Thread):
             details = {}
         if self.authid:
             details.setdefault('authid', self.authid)
-        details.setdefault('agent', u'swampyer-1.0')
-        details.setdefault('authmethods', self.authmethods or [u'anonymous'])
+        details.setdefault('agent', 'swampyer-1.0')
+        details.setdefault('authmethods', self.authmethods or ['anonymous'])
         details.setdefault('roles', {
                                         'subscriber': {},
                                         'publisher': {},
@@ -314,7 +368,7 @@ class WAMPClient(threading.Thread):
         """ Sends a RPC request to the WAMP server
         """
         if self._state == STATE_DISCONNECTED:
-            raise Exception("WAMP is currently disconnected!")
+            raise ExWAMPConnectionError("WAMP is currently disconnected!")
         options = {
             'disclose_me': True
         }
@@ -331,10 +385,10 @@ class WAMPClient(threading.Thread):
 
         if message == WAMP_ERROR:
             if message.args:
-                err = message.args[0]
+                err = message.args
             else:
-                err = message.error
-            raise ExInvocationError(err)
+                err = [message.error]
+            raise ExInvocationError(*err)
 
         return message
 
@@ -343,27 +397,33 @@ class WAMPClient(threading.Thread):
             for a response here. Just fire out a message
         """
         if self._state == STATE_DISCONNECTED:
-            raise Exception("WAMP is currently disconnected!")
+            raise ExWAMPConnectionError("WAMP is currently disconnected!")
         message = message.as_str()
         logger.debug("SND>: {}".format(message))
         if not self.ws:
-            raise Exception("WAMP is currently disconnected!")
-        self.ws.send(message)
+            raise ExWAMPConnectionError("WAMP is currently disconnected!")
+        try:
+            self.ws.send(message)
+        except websocket.WebSocketConnectionClosedException:
+            raise ExWAMPConnectionError("WAMP is currently disconnected!")
 
     def send_and_await_response(self,request):
         """ Used by most things. Sends out a request then awaits a response
             keyed by the request_id
         """
         if self._state == STATE_DISCONNECTED:
-            raise Exception("WAMP is currently disconnected!")
+            raise ExWAMPConnectionError("WAMP is currently disconnected!")
         wait_queue = queue.Queue()
         request_id = request.request_id
         self._requests_pending[request_id] = wait_queue;
         self.send_message(request)
         try:
-            return wait_queue.get(block=True,timeout=self.timeout)
-        except Exception as ex:
+            res = wait_queue.get(block=True,timeout=self.timeout)
+        except queue.Empty as ex:
             raise Exception("Did not receive a response!")
+        if isinstance(res, GOODBYE):
+            raise ExWAMPConnectionError("WAMP is currently disconnected!")
+        return res
 
     def dispatch_to_awaiting(self,result):
         """ Send dat ato the appropriate queues
@@ -403,11 +463,6 @@ class WAMPClient(threading.Thread):
         """
         pass
 
-    def handle_result(self, result):
-        """ Dispatch the result back to the appropriate awaiter
-        """
-        self.dispatch_to_awaiting(result)
-
     def handle_subscribed(self, result):
         """ Handle the successful subscription
         """
@@ -470,13 +525,24 @@ class WAMPClient(threading.Thread):
         """
         full_topic = self.get_full_uri(topic)
         result = self.send_and_await_response(SUBSCRIBE(
-                                    options={},
+                                    options=options or {},
                                     topic=full_topic
                                 ))
         if result == WAMP_SUBSCRIBED:
             if not callback:
                 callback = lambda a: None
             self._subscriptions[result.subscription_id] = [topic,callback]
+        return result
+
+    def unsubscribe(self, subscription_id):
+        """ Unsubscribe an existing subscription
+        """
+        result = self.send_and_await_response(UNSUBSCRIBE(subscription_id=subscription_id))
+        try:
+            del self._subscriptions[subscription_id]
+        except IndexError:
+            logger.warn("Subscription ID '%s' not found in local subscription list. Sent unsubscribe to router anyway.")
+        return result
 
     def publish(self,topic,options=None,args=None,kwargs=None):
         """ Publishes a messages to the server
@@ -501,7 +567,7 @@ class WAMPClient(threading.Thread):
                         kwargs=kwargs or {}
                       )
             self.send_message(request)
-            return request.request_id
+            return None
 
     def disconnect(self):
         """ Disconnect from the websocket and pause the process
@@ -514,6 +580,7 @@ class WAMPClient(threading.Thread):
             try:
                 if self._state == STATE_CONNECTED:
                     self.handle_leave()
+                    self.stop_heartbeat()
                     self.send_message(GOODBYE(
                           details={},
                           reason="wamp.error.system_shutdown"
@@ -543,8 +610,11 @@ class WAMPClient(threading.Thread):
                           reason="wamp.error.system_shutdown"
                         ))
         self._requests_pending = {}
+        self._last_ping_time = None
+        self._last_pong_time = None
         self.handle_disconnect()
 
+
     def shutdown(self):
         """ Request the system to shutdown the main loop and shutdown the system
             This is a one-way trip! Reconnecting requires a new connection
@@ -556,10 +626,10 @@ class WAMPClient(threading.Thread):
                 break
             time.sleep(0.1)
 
-    def start(self):
+    def start(self, **options):
         """ Initialize websockets, say hello, and start listening for events
         """
-        self.connect()
+        self.connect(**options)
         if not self.isAlive():
             super(WAMPClient,self).start()
         self.hello()
@@ -578,6 +648,19 @@ class WAMPClient(threading.Thread):
 
         return self
 
+    def unregister(self, registration_id):
+        result = self.send_and_await_response(UNREGISTER(registration_id=registration_id))
+        if result == WAMP_UNREGISTERED:
+            del self._registered_calls[result.registration_id]
+        elif result == WAMP_ERROR:
+            if result.args:
+                err = result.args
+            else:
+                err = [result.error]
+            raise ExInvocationError(*err)
+
+        return result
+
     def register(self,uri,callback,details=None):
         full_uri = self.get_full_uri(uri)
         result = self.send_and_await_response(REGISTER(
@@ -586,6 +669,13 @@ class WAMPClient(threading.Thread):
                   ))
         if result == WAMP_REGISTERED:
             self._registered_calls[result.registration_id] = [ uri, callback ]
+        elif result == WAMP_ERROR:
+            if result.args:
+                err = result.args
+            else:
+                err = [result.error]
+            raise ExInvocationError(*err)
+
         return result
 
     def run(self):
@@ -598,13 +688,17 @@ class WAMPClient(threading.Thread):
             # Find out if we have any data pending from the
             # server
             try:
+                if self._last_pong_time:
+                    since_last_pong = time.time() - self._last_pong_time
+                else:
+                    since_last_pong = None
 
                 # If we've been asked to stop running the
                 # request loop. We'll just sit and wait
                 # till we get asked to run again
                 if self._state not in [STATE_AUTHENTICATING,STATE_WEBSOCKET_CONNECTED,STATE_CONNECTED]:
                     self._request_loop_notify_restart.acquire()
-                    self._request_loop_notify_restart.wait(self._loop_timeout)
+                    self._request_loop_notify_restart.wait(self.loop_timeout)
                     self._request_loop_notify_restart.release()
                     continue
 
@@ -615,14 +709,37 @@ class WAMPClient(threading.Thread):
                     self._state = STATE_DISCONNECTED
                     continue
 
+                if since_last_pong and since_last_pong > self.heartbeat_timeout:
+                    # If the last ping response happened too long
+                    # ago, consider it a websocket timeout and
+                    # handle disconnect.
+                        raise ExWAMPConnectionError("Maximum websocket response delay of %s secs exceeded.", self.heartbeat_timeout)
+
                 # Okay, we think we're okay so let's try and read some data
-                data = self.ws.recv()
+                opcode, data = self.ws.recv_data(control_frame=True)
+                if opcode == websocket.ABNF.OPCODE_TEXT:
+                    # Try to decode the data as a utf-8 string. Replace any inconvertible characters
+                    # to the unicode `\uFFFD` character
+                    data = data.decode('utf-8', 'replace')
+
+                if opcode == websocket.ABNF.OPCODE_PONG:
+                    duration = time.time() - float(data)
+                    self._last_pong_time = time.time()
+                    data = None
+                    opcode = None
+                    logger.debug('Received websocket ping response in %s seconds', round(duration, 3))
+                    continue
+                
+                if opcode not in (websocket.ABNF.OPCODE_TEXT, websocket.ABNF.OPCODE_BINARY):
+                    continue
             except io.BlockingIOError:
                 continue
             except websocket.WebSocketTimeoutException:
                 continue
-            except websocket.WebSocketConnectionClosedException as ex:
-                logger.debug("WebSocketConnectionClosedException: Requesting disconnect:".format(ex))
+            except (websocket.WebSocketConnectionClosedException, ExWAMPConnectionError, socket.error) as ex:
+                logger.debug("WebSocket Exception. Requesting disconnect:".format(ex))
+                self._state = STATE_DISCONNECTED
+                self.stop_heartbeat()
                 self.disconnect()
 
                 # If the server disconnected, let's try and reconnect
@@ -655,7 +772,7 @@ class WAMPClient(threading.Thread):
                 except AttributeError as ex:
                     self.handle_unknown(message)
             except Exception as ex:
-                logger.error("ERROR in main loop: {}".format(ex))
+                logger.error("ERROR in main loop when receiving: {}".format(ex))
 
 class WAMPClientTicket(WAMPClient):
     username = None
@@ -669,7 +786,7 @@ class WAMPClientTicket(WAMPClient):
                 ):
 
         if not kwargs.get('authmethods'):
-            kwargs['authmethods'] = [u'ticket']
+            kwargs['authmethods'] = ['ticket']
         super(WAMPClientTicket,self).__init__(**kwargs)
         self.daemon = True
 
